use super::Engine;
use crate::{FileDescriptor, decrypt_deflate, encrypt_compress, salt};

use pathbufd::PathBufD;
use reqwest::{Client, StatusCode};
use std::{
    collections::HashMap,
    fs::write,
    io::{Error, Result},
    time::Duration,
};
use tokio::task::JoinHandle;

pub struct RentryEngine {
    pub http: Client,
    pub csrf_token: Option<String>,
}

impl RentryEngine {
    pub fn new() -> Self {
        Self {
            http: Client::new(),
            csrf_token: None,
        }
    }

    pub async fn get_csrf(&self, url: &str) -> String {
        println!("start: extract token {}", url);
        let body = self
            .http
            .get(url)
            .send()
            .await
            .unwrap()
            .text()
            .await
            .unwrap();

        let token = body
            .split("name=\"csrfmiddlewaretoken\" value=\"")
            .skip(1)
            .next()
            .unwrap()
            .split("\"")
            .next()
            .unwrap()
            .to_string();

        println!("start: token obtained");
        token
    }
}

impl Engine for RentryEngine {
    const CHUNK_SIZE: usize = 150_000; // chunks are given 8 bytes for extra characters

    async fn auth(&mut self) -> Result<()> {
        self.csrf_token = Some(self.get_csrf("https://rentry.co").await);
        Ok(())
    }

    async fn process(&self, name: String, data: Vec<u8>) -> Result<FileDescriptor> {
        let token = if let Some(ref token) = self.csrf_token {
            token.to_owned()
        } else {
            return Err(Error::other("no csrf token extracted"));
        };

        let (seed, key, data) = encrypt_compress(data);
        let mut descriptor = FileDescriptor {
            name,
            key,
            seed,
            engine_data: {
                let mut data = HashMap::new();
                data.insert("token".to_string(), token);
                data
            },
            chunks: Vec::new(),
        };

        let chars = data.chars().collect::<Vec<char>>();
        let chunks = chars.chunks(Self::CHUNK_SIZE);
        let chunks_num = chunks.len();

        let mut handles: Vec<JoinHandle<()>> = Vec::new();

        for (i, chunk) in chunks.enumerate() {
            let id = salt();
            println!("cnksv: start chunk save ({}/{chunks_num})", i + 1);

            // spawn worker thread
            let c = chunk.to_vec();

            let mut worker_engine = Self::new();
            worker_engine.csrf_token = self.csrf_token.clone();

            let id_c = id.clone();
            tokio::time::sleep(Duration::from_millis(750)).await; // sleep to avoid rate limit

            let handle = tokio::spawn(async move {
                let mut b = worker_engine
                    .create_chunk(&id_c, c.iter().collect())
                    .await
                    .expect("failed to create chunk");

                while b.contains("e436") {
                    println!(
                        "\x1b[91mcnksv: upload failed (hit limit on {})\x1b[0m",
                        i + 1
                    );
                    println!("cnksv: creation limit reached, waiting to try again");

                    tokio::time::sleep(Duration::from_millis(10000)).await;

                    b = worker_engine
                        .create_chunk(&id_c, c.iter().collect())
                        .await
                        .expect("failed to create chunk");
                }
            });

            handles.push(handle);
            descriptor.chunks.push(id.clone());
        }

        // wait for all handles to finish
        println!("cnksv: waiting for all handles to finish");
        loop {
            let mut all_finished = true;

            for handle in &handles {
                if !handle.is_finished() {
                    all_finished = false;
                    break;
                }
            }

            if all_finished {
                break;
            }

            // sleep a little to save cpu
            tokio::time::sleep(Duration::from_millis(150)).await;
        }

        // ...
        Ok(descriptor)
    }

    async fn reconstruct(&self, descriptor: FileDescriptor) -> Result<()> {
        let mut encoded_string: String = String::new();

        let chunks_num = descriptor.chunks.len();
        for (i, chunk) in descriptor.chunks.iter().enumerate() {
            tokio::time::sleep(Duration::from_millis(250)).await;
            println!("cnkrd: start chunk read ({}/{chunks_num})", i + 1);

            if let Ok(d) = self.get_chunk(&chunk).await {
                println!("cnkrd: read {} bytes", d.len());
                encoded_string += &d;
            } else {
                println!("cnkrd: waiting to retry chunk read");
                tokio::time::sleep(Duration::from_millis(1000)).await;
                encoded_string += &self.get_chunk(&chunk).await?;
            }
        }

        let decoded = decrypt_deflate(descriptor.seed, descriptor.key, encoded_string);
        write(PathBufD::current().join(descriptor.name), decoded)?;

        Ok(())
    }

    async fn delete(&mut self, descriptor: FileDescriptor) -> Result<()> {
        self.csrf_token = Some(
            descriptor
                .engine_data
                .get("token")
                .expect("rentry engine requires engine_data.token")
                .to_owned(),
        );

        for (i, chunk) in descriptor.chunks.iter().enumerate() {
            println!(
                "cnkde: start chunk delete ({}/{})",
                i + 1,
                descriptor.chunks.len()
            );

            if self.delete_chunk(&chunk).await.is_err() {
                // wait and then retry
                println!("cnkde: waiting to retry chunk deletion");
                tokio::time::sleep(Duration::from_millis(750)).await;

                if self.delete_chunk(&chunk).await.is_err() {
                    println!("cnkde: failed to delete chunk {} (skipping)", i + 1);
                    continue;
                };
            }
        }

        Ok(())
    }

    async fn get_chunk(&self, id: &str) -> Result<String> {
        println!("cnkrq: chunk request");
        let res = self
            .http
            .get(format!("https://rentry.co/renbin_{id}"))
            .send()
            .await
            .unwrap();

        if res.status() != StatusCode::OK {
            return Err(Error::other("remote error"));
        }

        let body = res.text().await.unwrap();

        // this shows how useful that raw key thing is for you
        let content = body
            .split("<div><p>")
            .skip(1)
            .next()
            .unwrap()
            .split("</p></div>")
            .next()
            .unwrap()
            .to_string();

        println!("cnkrd: chunk read");
        Ok(content)
    }

    async fn create_chunk(&self, id: &str, data: String) -> Result<String> {
        let token = if let Some(ref token) = self.csrf_token {
            token
        } else {
            return Err(Error::other("no csrf token extracted"));
        };

        // create body
        let mut body_map: HashMap<&str, String> = HashMap::new();
        body_map.insert("csrfmiddlewaretoken", token.to_owned());
        body_map.insert("text", data);
        body_map.insert("metadata", String::new());
        body_map.insert("edit_code", token.to_owned());
        body_map.insert("url", format!("renbin_{id}"));

        // ...
        let res = self
            .http
            .post("https://rentry.co")
            .form(&body_map)
            .header("Cookie", format!("csrftoken={token}"))
            .header("Referer", "https://rentry.co/")
            .send()
            .await
            .expect("failed to create chunk");

        if res.status() != StatusCode::OK {
            return Err(Error::other("remote error"));
        }

        println!("cnksv: chunk saved");
        Ok(res.text().await.unwrap())
    }

    async fn delete_chunk(&self, id: &str) -> Result<()> {
        let token = if let Some(ref token) = self.csrf_token {
            token
        } else {
            return Err(Error::other("no csrf token extracted"));
        };

        // create body
        let mut body_map: HashMap<&str, String> = HashMap::new();
        body_map.insert("csrfmiddlewaretoken", token.to_owned());
        body_map.insert("text", "renbin_delete".to_string());
        body_map.insert("metadata", String::new());
        body_map.insert("edit_code", token.to_owned());
        body_map.insert("new_edit_code", String::new());
        body_map.insert("new_url", String::new());
        body_map.insert("new_modify_code", String::new());
        body_map.insert("delete", "delete".to_string());

        // the rentry api is one of the worst i've ever seen
        let url = format!("https://rentry.co/renbin_{id}/edit");

        let res = self
            .http
            .post(&url)
            .form(&body_map)
            .header("Cookie", format!("csrftoken={token}"))
            .header("Referer", &url)
            .send()
            .await
            .expect("failed to delete chunk");

        if res.status() != StatusCode::OK {
            return Err(Error::other("remote error"));
        }

        println!("cnkde: chunk deleted");
        Ok(())
    }
}
