use crate::FileDescriptor;
use pathbufd::PathBufD;
use serde::{Deserialize, Serialize};
use std::{
    fs::{metadata, read_dir, read_to_string, write},
    io::Result,
};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct Ledger {
    pub name: String,
    pub dirs: Vec<String>,
    pub files: Vec<FileDescriptor>,
}

impl Ledger {
    /// Read a [`Ledger`] from the given `path`.
    pub fn read(path: PathBufD) -> Self {
        toml::from_str(&read_to_string(path).expect("failed to read file"))
            .expect("failed to deserialize file")
    }

    /// Write a [`Ledger`] into the given `path`.
    pub fn write(&self, path: PathBufD) -> Result<()> {
        write(
            path,
            toml::to_string_pretty(&self).expect("failed to serialize file"),
        )
    }

    /// Recursively collect paths from a directory.
    ///
    /// # Returns
    /// `(files, dirs)`
    pub fn collect_paths(root: &PathBufD, skip: &Vec<String>) -> (Vec<PathBufD>, Vec<String>) {
        let mut dirs = Vec::new();
        let mut files = Vec::new();

        for x in read_dir(&root).expect("failed to read dir") {
            let x = x.unwrap();
            let path = root.join(x.file_name());

            if skip.contains(&x.file_name().into_string().unwrap()) {
                continue;
            }

            let stat = metadata(&path).expect("failed to stat file");

            if stat.is_dir() {
                dirs.push(x.path().to_str().unwrap().replace("../", ""));

                let (files_, dirs_) = Self::collect_paths(&path, skip);

                for file in files_ {
                    files.push(file);
                }

                for dir in dirs_ {
                    dirs.push(dir);
                }
            } else {
                files.push(x.path().into());
            }
        }

        (files, dirs)
    }
}
