pub mod engine;
pub mod ledger;

use aes_gcm::{
    Aes256Gcm, Nonce,
    aead::{AeadCore, AeadInPlace, KeyInit, OsRng},
    aes::cipher::typenum,
};
use base64::{Engine, engine::general_purpose::STANDARD};
use flate2::{Compression, read::GzDecoder, write::GzEncoder};
use pathbufd::PathBufD as PathBuf;
use rand::{Rng, distr::Alphanumeric, rng};
use serde::{Deserialize, Serialize};
use std::{
    collections::HashMap,
    fs::{read_to_string, write},
    io::{Result, prelude::*},
};

#[derive(Clone, Debug, Serialize, Deserialize)]
pub struct FileDescriptor {
    /// The name of the file.
    pub name: String,
    /// The encryption key used for the file.
    pub key: String,
    /// The seed (nonce) used for the file.
    pub seed: String,
    /// Data needed by the engine to read this file.
    pub engine_data: HashMap<String, String>,
    /// A list of all locations to retrieve chunks from.
    /// Each chunk **must** be in the correct order.
    pub chunks: Vec<String>,
}

impl FileDescriptor {
    /// Read a [`FileDescriptor`] from the given `path`.
    pub fn read(path: PathBuf) -> Self {
        toml::from_str(&read_to_string(path).expect("failed to read file"))
            .expect("failed to deserialize file")
    }

    /// Write a [`FileDescriptor`] into the given `path`.
    pub fn write(&self, path: PathBuf) -> Result<()> {
        write(
            path,
            toml::to_string_pretty(&self).expect("failed to serialize file"),
        )
    }
}

/// Create a random (not necessarily unique) 16-byte long string.
pub fn salt() -> String {
    rng()
        .sample_iter(&Alphanumeric)
        .take(16)
        .map(char::from)
        .collect()
}

/// Encrypt and compress the given `data` using the given `key`.
///
/// # Returns
/// `(nonce, key, encrypted as base64)`
pub fn encrypt_compress(data: Vec<u8>) -> (String, String, String) {
    // compress
    let mut encoder = GzEncoder::new(Vec::new(), Compression::default());
    encoder.write_all(&data).expect("failed to compress");
    println!("encod: data compressed");

    let mut compressed_buffer: Vec<u8> = encoder.finish().unwrap();

    // encrypt
    let key = Aes256Gcm::generate_key(&mut OsRng);
    let cipher = Aes256Gcm::new(&key);
    let nonce = Aes256Gcm::generate_nonce(&mut OsRng);

    cipher
        .encrypt_in_place(&nonce, b"", &mut compressed_buffer)
        .expect("failed to encrypt");

    println!("encod: data encrypted");

    // return
    (
        STANDARD.encode(nonce.to_vec()),
        STANDARD.encode(key.to_vec()),
        STANDARD.encode(compressed_buffer.to_vec()),
    )
}

/// Decrypt and uncompress the given `data` using the given `key`.
///
/// # Returns
/// `(key, encrypted)`
pub fn decrypt_deflate(nonce: String, key: String, data: String) -> Vec<u8> {
    // decrypt
    // decryption must happen first since we must undo what was done previously
    let key = STANDARD.decode(key).expect("recieved invalid base64 key");
    let cipher = Aes256Gcm::new_from_slice(&key.as_slice()).unwrap();
    let nonce = Nonce::<typenum::U12>::clone_from_slice(
        STANDARD
            .decode(nonce)
            .expect("received invalid base64 nonce")
            .as_slice(),
    );

    let data = STANDARD.decode(data).expect("received invalid base64 data");
    let mut decrypted_buffer: Vec<u8> = data;

    cipher
        .decrypt_in_place(&nonce, b"", &mut decrypted_buffer)
        .expect("failed to decrypt");

    println!("decod: data decrypted");

    // compress
    let mut decoder = GzDecoder::new(decrypted_buffer.as_slice());

    let mut decompressed_buffer: Vec<u8> = Vec::new();
    decoder
        .read_to_end(&mut decompressed_buffer)
        .expect("failed to decompress");

    println!("decod: data decompressed");

    // return
    decompressed_buffer
}
