extern crate renbin;

use clap::Parser;
use pathbufd::PathBufD;
use renbin::{
    FileDescriptor,
    engine::{Engine, fs::FsEngine, rentry::RentryEngine},
    ledger::Ledger,
};
use std::{
    fs::{create_dir, read, remove_file},
    path::PathBuf,
    str::FromStr,
};

#[derive(Parser, Debug)]
#[command(version, about, long_about = None)]
struct Args {
    #[arg(short = 'd', long = "decode", action)]
    decode: bool,
    #[arg(short = 'x', long = "delete", action)]
    delete: bool,
    #[arg(short = 'r', long = "recursive", action)]
    recursive: bool,
    #[arg(short = 's', long = "skip", action)]
    skip: Vec<String>,
    #[arg(short = 'n', long = "name")]
    name: Option<String>,
    #[arg(short = 'i', long = "input")]
    input: String,
    #[arg(short = 'e', long = "engine")]
    #[clap(default_value = "fs")]
    engine: String,
}

#[tokio::main]
async fn main() {
    let args = Args::parse();

    if args.recursive
        && let Some(name) = args.name
        && !args.decode
    {
        // build ledger
        let pathbuf = PathBuf::from_str(&args.input).unwrap();
        let (files, dirs) = Ledger::collect_paths(&pathbuf.clone().into(), &args.skip);

        Ledger {
            name: name.clone(),
            dirs,
            files: {
                let mut descriptors = Vec::new();

                if args.engine == "fs" {
                    for file in files {
                        let descriptor = FsEngine
                            .process(
                                file.as_path()
                                    .to_str()
                                    .unwrap()
                                    .to_string()
                                    .replace("../", ""),
                                read(file).expect("failed to read file"),
                            )
                            .await
                            .expect("failed to process file");

                        descriptors.push(descriptor);
                    }
                } else if args.engine == "rentry" {
                    let mut engine = RentryEngine::new();
                    engine.auth().await.expect("failed to extract csrf token");

                    for file in files {
                        let descriptor = engine
                            .process(
                                file.as_path()
                                    .to_str()
                                    .unwrap()
                                    .to_string()
                                    .replace("../", ""),
                                read(file).expect("failed to read file"),
                            )
                            .await
                            .expect("failed to process file");

                        descriptors.push(descriptor);
                    }
                } else {
                    unreachable!("invalid engine")
                };

                descriptors
            },
        }
        .write(PathBuf::from_str(&format!("{}.toml", name)).unwrap().into())
        .expect("failed to write ledger");

        return;
    } else if args.recursive && args.decode {
        // decode ledger
        let path = PathBuf::from(args.input);
        let ledger = Ledger::read(path.clone().into());

        // create dirs
        for dir in ledger.dirs {
            create_dir(dir).expect("failed to create dir");
        }

        // reconstruct files
        for file in ledger.files {
            if args.engine == "fs" {
                FsEngine
                    .reconstruct(file)
                    .await
                    .expect("failed to reconstruct file");
            } else if args.engine == "rentry" {
                let mut engine = RentryEngine::new();
                engine.auth().await.expect("failed to extract csrf token");

                engine
                    .reconstruct(file)
                    .await
                    .expect("failed to reconstruct file");
            } else {
                panic!("unknown engine type");
            };
        }

        // ...
        return;
    } else if args.recursive && args.delete {
        // delete ledger
        let path = PathBuf::from(args.input);
        let ledger = Ledger::read(path.clone().into());

        if args.engine == "fs" {
            for file in ledger.files {
                FsEngine
                    .delete(file)
                    .await
                    .expect("failed to reconstruct file");
            }
        } else if args.engine == "rentry" {
            let mut engine = RentryEngine::new();
            engine.auth().await.expect("failed to extract csrf token");

            for file in ledger.files {
                engine.delete(file).await.expect("failed to delete file");
            }
        } else {
            panic!("unknown engine type");
        };

        remove_file(path).expect("failed to delete file descriptor");
        return;
    }

    if args.delete {
        // delete
        let path = PathBuf::from(args.input);

        if args.engine == "fs" {
            FsEngine
                .delete(FileDescriptor::read(path.clone().into()))
                .await
                .expect("failed to reconstruct file");
        } else if args.engine == "rentry" {
            let mut engine = RentryEngine::new();
            engine.auth().await.expect("failed to extract csrf token");

            engine
                .delete(FileDescriptor::read(path.clone().into()))
                .await
                .expect("failed to delete file");
        } else {
            panic!("unknown engine type");
        };

        remove_file(path).expect("failed to delete file descriptor");
        return;
    }

    if !args.decode {
        // encode
        let pathbuf = PathBuf::from(args.input);

        if args.engine == "fs" {
            let descriptor = FsEngine
                .process(
                    pathbuf.file_name().unwrap().to_str().unwrap().to_string(),
                    read(pathbuf).expect("failed to read file"),
                )
                .await
                .expect("failed to process file");

            descriptor
                .write(PathBufD::current().join(format!("{}.toml", descriptor.name)))
                .expect("failed to write descriptor");
        } else if args.engine == "rentry" {
            let mut engine = RentryEngine::new();
            engine.auth().await.expect("failed to extract csrf token");

            let descriptor = engine
                .process(
                    pathbuf.file_name().unwrap().to_str().unwrap().to_string(),
                    read(pathbuf).expect("failed to read file"),
                )
                .await
                .expect("failed to process file");

            descriptor
                .write(PathBufD::current().join(format!("{}.toml", descriptor.name)))
                .expect("failed to write descriptor");
        } else {
            panic!("unknown engine type");
        };
    } else {
        // decode
        if args.engine == "fs" {
            FsEngine
                .reconstruct(FileDescriptor::read(PathBuf::from(args.input).into()))
                .await
                .expect("failed to reconstruct file");
        } else if args.engine == "rentry" {
            let mut engine = RentryEngine::new();
            engine.auth().await.expect("failed to extract csrf token");

            engine
                .reconstruct(FileDescriptor::read(PathBuf::from(args.input).into()))
                .await
                .expect("failed to reconstruct file");
        } else {
            panic!("unknown engine type");
        };
    }
}
